#pragma once

#include "../HashFunction.h"
#include "CuckooHashTableView.h"

#include <TNL/Containers/Array.h>

#include <cstdint>
#include <iostream>
#include <functional>
#include <vector>

using namespace TNL::Containers;


/** CuckooHashTable - HashMap for template types
 * 
 * K - type of keys
 * V - type of values
 */
template<typename Item, typename Key, typename Device>
class CuckooHashTable {
    using Self = CuckooHashTable<Item, Key, Device>;
    using ViewType = CuckooHashTableView<Item, Key, Device>;
    using Entry = typename ViewType::Entry;
    friend ViewType::CuckooHashTableView(Self&, const typename Array<Item, Device>::ConstViewType);

  public:
    ///< Reports the number of performed attempts to build the table
    int generations() const;

    ///< Reports the number of attempts to insert duplicates while building the table
    int duplicates() const;

    void debug_print() {
      for (int i = 0; i < m_table.getSize(); i++) {
        auto entry = m_table.getElement(i);
        if (entry >= 0)
          std::cout << i << ' ' << m_content.getElement(entry).item.key /*<< ' ' << m_content.getElement(entry.item).value*/ << std::endl;
        else
          std::cout << i << ' ' << "Not occupied" << std::endl;
      }
    }


  protected:
    ///< Copy constructor
    CuckooHashTable(const CuckooHashTable<Item, Key, Device>& other);

    ///< Move constructor
    CuckooHashTable(CuckooHashTable<Item, Key, Device>&& other);

    /**Constructor
     * values - vector of <key, value> pairs to populate the table
     * nHashFunctions = number of hash functions to generate and use
     */
    CuckooHashTable(const Array<Item, Device>& values, uint8_t nHashFunctions);
    
    /**Constructor
     * 
     * tableSize - number of rows in the table
     * nHashFunctions - number of hash functions to generate and use
     * nIters - maximal number of insertion attempts before starting over with regenerated hash functions
     * values - vector of <key, value> pairs to populate the table
     */
    CuckooHashTable(int tableSize, uint8_t nHashFunctions, int nIters, 
                    const Array<Item, Device>& values = {});

    ///< Copy assignment
    CuckooHashTable<Item, Key, Device>& operator=(const CuckooHashTable<Item, Key, Device>& other);

    Array<HashFunction<Key>, Device> m_hashFunctions;
    Array<int, Device> m_table;
    Array<Entry, Device> m_content;

  private:
    int m_iterations;
  
  protected:
    std::shared_ptr<ViewType> m_view;
};


#include "CuckooHashTable.hpp"
