#pragma once

#include "CuckooHashTableView.h"

#include <TNL/Algorithms/ParallelFor.h>
#include <TNL/Allocators/Default.h>
#include <TNL/Exceptions/CudaRuntimeError.h>


template<typename Item, typename Key, typename Device>
CuckooHashTableView<Item, Key, Device>::CuckooHashTableView(TableType& table, const typename Array<Item, Device>::ConstViewType items):
        m_hashFunctions(table.m_hashFunctions.getView()),
        m_table(table.m_table.getView()),
        m_content(table.m_content.getView()),
        m_iterations(table.m_iterations),
        m_generations(0),
        m_duplicates(0) {
    m_table.setValue(-1);
    if (!items.empty()) {
        copy_items(items);
        build();
    }
}

template<typename Item, typename Key, typename Device>
void CuckooHashTableView<Item, Key, Device>::init_hash_functions() {
    auto hf = m_hashFunctions;
    auto _init = [hf] __cuda_callable__ (int i, int rand, int max) mutable {
        RandomIntGenerator r(rand + i);
        hf[i].init(r, max);
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, m_hashFunctions.getSize(), _init, m_rand(), m_table.getSize());
    m_generations++;
}

template<typename Item, typename Key, typename Device>
__cuda_callable__ bool insert_impl(int entry, 
                                   ArrayView<typename CuckooHashTableView<Item, Key, Device>::Entry, Device> items, 
                                   bool& duplicate, int iter,
                                   ArrayView<HashFunction<Key>, Device> hf, 
                                   ArrayView<int, Device> table) {
    duplicate = false;
    int index = hf[0](items[entry].item.key);

    for (int i = 0; i < iter; i++) {
        #ifdef __CUDA_ARCH__
            entry = atomicExch(&table[index], entry);
        #else
            std::swap(table[index], entry);
        #endif

        // If candidate was free, stop
        if (entry < 0)
            return true;

        // Don't allow inserting duplicates
        if (items[entry].item.key == items[table[index]].item.key) {
            duplicate = true;
            return false;
        }

        // Find a new position for the old content
        int newHash = (items[entry].hashFunction + 1) % hf.getSize();
        items[entry].hashFunction = newHash;
        index = hf[newHash](items[entry].item.key); 
    }

    // maximum number of iterations reached
    return false;
}

template<typename Item, typename Key, typename Device>
void CuckooHashTableView<Item, Key, Device>::copy_items(const typename Array<Item, Device>::ConstViewType items) {
    auto _init_content = [items] __cuda_callable__ (int i, ArrayView<Entry, Device> content) mutable {
        content[i] = {items[i], -1};
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, items.getSize(), _init_content, m_content);
}

template<typename Item, typename Key, typename Device>
void CuckooHashTableView<Item, Key, Device>::build() {
    Array<unsigned, Device> result(2);
    auto rview = result.getView();
    rview.setValue(0);
    auto _build = [rview] __cuda_callable__ (int i, 
                                             decltype(m_content) items,
                                             decltype(m_iterations) iter,
                                             decltype(m_hashFunctions) hf,
                                             decltype(m_table) table) mutable {
        bool dup;
        if (!insert_impl<Item, Key, Device>(i, items, dup, iter, hf, table)) {
            if constexpr (std::is_same<Device, TNL::Devices::Cuda>::value)
                atomicAdd(&rview[!dup], 1);
            else
                rview[!dup]++;
        }
    };
    do {
        init_hash_functions();
        m_table.setValue(-1);
        m_duplicates = 0;
        rview.setValue(0);
        TNL::Algorithms::ParallelFor<Device>::exec(0, m_content.getSize(), _build, m_content, m_iterations, m_hashFunctions, m_table);
    } while (rview.getElement(1));
    m_duplicates = rview.getElement(0);
}

template<typename Item, typename Key, typename Device>
int CuckooHashTableView<Item, Key, Device>::generations() const { return m_generations; }

template<typename Item, typename Key, typename Device>
int CuckooHashTableView<Item, Key, Device>::duplicates() const { return m_duplicates; }

template<typename Item, typename Key, typename Device>
bool CuckooHashTableView<Item, Key, Device>::find(const Key& key, ArrayView<Item, Device> item) const {
    Array<int, Device> res(1);
    res.setValue(-1);
    auto rview = res.getView();
    auto _find = [rview, key, item] __cuda_callable__ (int i, 
                                                 ArrayView<HashFunction<Key>, Device> hf, 
                                                 ArrayView<int, Device> table, 
                                                 ArrayView<Entry, Device> cont) mutable { 
        int pos = hf[i](key);
        if (table[pos] >= 0 && cont[table[pos]].item.key == key) {
            rview[0] = pos;
            item[0] = cont[table[pos]].item;
        }
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, m_hashFunctions.getSize(), _find, m_hashFunctions, m_table, m_content);
    int result = rview.getElement(0);
    return result > -1;
}
