#pragma once

#include "HashGraphSView.h"

#include <TNL/Algorithms/ParallelFor.h>


template<typename Item, typename Key, typename Device>
HashGraphSView<Item, Key, Device>::HashGraphSView(TableType& table,
                                                    const typename Array<Item, Device>::ConstViewType items) :
            m_content(table.m_table.getView()),
            m_table(table.m_table.getView()),
            m_csr(table.m_csr.getView()),
            m_segmentsSizes(table.m_segmentSizes.getView()),
            m_hash(31, 10538, items.getSize()) {
    build();
}

template<typename Item, typename Key, typename Device>
void HashGraphSView<Item, Key, Device>::build() {
    auto hash = m_hash;
    auto init_hashes = [hash, hashes, input] __cuda_callable__ (int i) mutable {
        hashes[i] = hash(input[i].key);
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, m_items.getSize(), init_hashes);
    m_segmentsSizes.setValue(0);
    auto count_hashes = [hashes, counter] __cuda_callable__ (int i) mutable {
        #ifdef __CUDA_ARCH__
        atomicAdd(&counter[hashes[i]], 1);
        #else
        m_segmentSizes[hashes[i]]++;
        #endif
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, hashes.getSize(), count_hashes);
    fill_csr();
    counter.setValue(0, 0, counter.getSize());
    auto offset = m_offset;
    auto items = m_items;
    auto place = [offset, counter, hashes, items] __cuda_callable__ (int i) mutable {
        #ifdef __CUDA_ARCH__
        int pos = offset[hashes[i]] + atomicAdd(&counter[hashes[i]], 1);
        #else
        int pos = offset[hashes[i]] + counter[hashes[i]]++;
        #endif
        items[pos] = i;
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, counter.getSize(), place);
}

template<typename Item, typename Key, typename Device>
HashGraphSView<Item, Key, Device>::~HashGraphSView() {}

template<typename Item, typename Key, typename Device>
void HashGraphSView<Item, Key, Device>::fill_csr() {
    auto offset = m_offset;
    auto fill = [offset, counter] __cuda_callable__ (int i) mutable {
        offset[i] = i >= counter.getSize() ? 0 : counter[i];
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, offset.getSize(), fill);

    auto reduce = [offset] __cuda_callable__ (int k, int d) mutable {
        if (k % (1 << (d + 1)) == 0)
            offset[k + (1 << (d + 1)) - 1] = offset[k + (1 << d) - 1]
                                           + offset[k + (1 << (d + 1)) - 1];
    };
    for (int d = 0; d < log2(offset.getSize() - 1); d++)
        TNL::Algorithms::ParallelFor<Device>::exec(0, offset.getSize(), reduce, d);

    offset.setElement(offset.getSize() - 1, 0);
    auto up_sweep = [offset] __cuda_callable__ (int r, int d) mutable {
        if ((offset.getSize() - 1 - r) % (1 << d) == 0) {
            int l = r - (1 << (d - 1));
            int t = offset[r];
            offset[r] += offset[l];
            offset[l] = t;
        }
    };
    for (int d = log2(offset.getSize()); d > 0; d--)
        TNL::Algorithms::ParallelFor<Device>::exec(0, offset.getSize(), up_sweep, d);
}

template<typename Item, typename Key, typename Device>
int HashGraphSView<Item, Key, Device>::duplicates() const {
    return -1;
}

template<typename Item, typename Key, typename Device>
bool HashGraphSView<Item, Key, Device>::find(const Key& key, ArrayView<Item, Device> item) const {
    int hash = m_hash(key);
    int end = m_offset.getElement(hash + 1);
    auto content = m_content;
    Array<int, Device> result(1);
    auto rView = result.getView();
    rView.setValue(-1);
    auto items = m_items;
    auto _find = [content, items, item, rView, key] __cuda_callable__ (int i) mutable {
        if (content[items[i]].key == key) {
            rView[0] = items[i];
            item[0] = content[items[i]];
        }
    };
    TNL::Algorithms::ParallelFor<Device>::exec(m_offset.getElement(hash), end, _find);
    return rView.getElement(0) > -1
}
