#pragma once

#include "HashGraphV1View.h"

#include <TNL/Algorithms/ParallelFor.h>


template<typename Item, typename Key, typename Device>
HashGraphV1View<Item, Key, Device>::HashGraphV1View(TableType* table) :
            m_content(table->m_content.getView()),
            m_items(table->m_items.getView()),
            m_offset(table->m_offset.getView()),
            m_hash(31, 10538, m_items.getSize()) {
    build();
}

template<typename Item, typename Key, typename Device>
void HashGraphV1View<Item, Key, Device>::build() {
    auto hashArr = Array<int, Device>(m_content.getSize());
    auto hashes = hashArr.getView();
    auto input = m_content;
    auto hash = m_hash;
    auto init_hashes = [hash, hashes, input] __cuda_callable__ (int i) mutable {
        hashes[i] = hash(input[i].key);
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, m_items.getSize(), init_hashes);
    auto countArr = Array<int, Device>(m_content.getSize(), 0);
    auto counter = countArr.getView();
    auto count_hashes = [hashes, counter] __cuda_callable__ (int i) mutable {
        #ifdef __CUDA_ARCH__
        atomicAdd(&counter[hashes[i]], 1);
        #else
        counter[hashes[i]]++;
        #endif
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, hashes.getSize(), count_hashes);
    fill_offset(counter);
    counter.setValue(0);
    auto offset = m_offset;
    auto items = m_items;
    auto place = [offset, counter, hashes, items] __cuda_callable__ (int i) mutable {
        #ifdef __CUDA_ARCH__
        int pos = offset[hashes[i]] + atomicAdd(&counter[hashes[i]], 1);
        #else
        int pos = offset[hashes[i]] + counter[hashes[i]]++;
        #endif
        items[pos] = i;
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, counter.getSize(), place);
}

template<typename Item, typename Key, typename Device>
HashGraphV1View<Item, Key, Device>::~HashGraphV1View() {}

template<typename Item, typename Key, typename Device>
void HashGraphV1View<Item, Key, Device>::fill_offset(const ArrayView<int, Device>& counter) {
    auto offset = m_offset;
    auto fill = [offset, counter] __cuda_callable__ (int i) mutable {
        offset[i] = i >= counter.getSize() ? 0 : counter[i];
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, offset.getSize(), fill);

    auto reduce = [offset] __cuda_callable__ (int k, int d) mutable {
        if (k % (1 << (d + 1)) == 0)
            offset[k + (1 << (d + 1)) - 1] = offset[k + (1 << d) - 1]
                                           + offset[k + (1 << (d + 1)) - 1];
    };
    for (int d = 0; d < log2(offset.getSize() - 1); d++)
        TNL::Algorithms::ParallelFor<Device>::exec(0, offset.getSize() - 1, reduce, d);

    offset.setElement(offset.getSize() - 1, offset.getElement(offset.getSize() - 2));
    offset.setElement(offset.getSize() - 2, 0);
    auto up_sweep = [offset] __cuda_callable__ (int r, int d) mutable {
        if ((offset.getSize() - 2 - r) % (1 << d) == 0) {
            int l = r - (1 << (d - 1));
            int t = offset[r];
            offset[r] += offset[l];
            offset[l] = t;
        }
    };
    for (int d = log2(offset.getSize() - 1); d > 0; d--)
        TNL::Algorithms::ParallelFor<Device>::exec(0, offset.getSize() - 1, up_sweep, d);
}

template<typename Item, typename Key, typename Device>
int HashGraphV1View<Item, Key, Device>::duplicates() const {
    // TODO
    return -1;
}

template<typename Item, typename Key, typename Device>
bool HashGraphV1View<Item, Key, Device>::find(const Key& key, ArrayView<Item, Device> item) const {
    int hash = m_hash(key);
    int end = m_offset.getElement(hash + 1);
    auto content = m_content;
    Array<int, Device> result(1);
    auto rView = result.getView();
    rView.setValue(-1);
    auto items = m_items;
    auto _find = [content, items, item, rView, key] __cuda_callable__ (int i) mutable {
        if (content[items[i]].key == key) {
            rView[0] = items[i];
            item[0] = content[items[i]];
        }
    };
    TNL::Algorithms::ParallelFor<Device>::exec(m_offset.getElement(hash), end, _find);
    return rView.getElement(0) > -1;
}
