#pragma once

#include "HashGraphV2View.h"

#include <TNL/Algorithms/ParallelFor.h>


template<typename Item, typename Key, typename Device>
HashGraphV2View<Item, Key, Device>::HashGraphV2View(TableType* table) :
            m_content(table->m_content.getView()),
            m_items(table->m_items.getView()),
            m_offset(table->m_offset.getView()),
            m_hash(31, 10538, m_content.getSize()) {
    build();
}

template<typename Item, typename Key, typename Device>
void HashGraphV2View<Item, Key, Device>::build() {
    // Phase 1
    int nBins = int(ceil(double(m_content.getSize()) / 16));
    int binSize = int(ceil(double(m_content.getSize()) / nBins));
    int tableSize = nBins * binSize;
    Array<int, Device> bcArr(nBins, 0);
    auto binCounter = bcArr.getView();
    auto content = m_content;
    auto hash = m_hash;
    Array<int, Device> hArr(m_content.getSize());
    auto hashes = hArr.getView();
    TNL::Algorithms::ParallelFor<Device>::exec(0, m_content.getSize(), [hashes, hash, content] __cuda_callable__ (int i) mutable {
        hashes[i] = hash(content[i].key);
    });
    TNL::Algorithms::ParallelFor<Device>::exec(0, m_content.getSize(), [hashes, binSize, binCounter] __cuda_callable__ (int i) mutable {
        int bin = hashes[i] / binSize;
        #ifdef __CUDA_ARCH__
        atomicAdd(&binCounter[bin], 1);
        #else
        binCounter[bin]++;
        #endif
    });
    // Phase 2
    Array<int, Device> boArr((1 << int(ceil(log2(nBins)))) + 1, 0);
    auto binOffset = boArr.getView();
    fill_offset(binCounter, binOffset); 
    binCounter.setValue(0);
    Array<int, Device> reArr(m_content.getSize());
    auto reorg = reArr.getView();
    TNL::Algorithms::ParallelFor<Device>::exec(0, m_content.getSize(),
                                               [hashes, binSize, binCounter, binOffset, reorg]
                                               __cuda_callable__ (int i) mutable {
        int bin = hashes[i] / binSize;
        #ifdef __CUDA_ARCH__
        int pos = atomicAdd(&binCounter[bin], 1) + binOffset[bin];
        atomicExch(&reorg[pos], i);
        #else
        int pos = binCounter[bin]++ + binOffset[bin];
        reorg[pos] = i;
        #endif
    });
    // Phase 3
    Array<int, Device> cnArr(m_content.getSize(), 0);
    auto counter = cnArr.getView();
    TNL::Algorithms::ParallelFor<Device>::exec(0, m_content.getSize(), [hashes, reorg, counter] __cuda_callable__ (int i) mutable {
        int pos = hashes[reorg[i]];
        #ifdef __CUDA_ARCH__
        atomicAdd(&counter[pos], 1);
        #else
        counter[pos]++;
        #endif
    });
    Array<int, Device> ofArr();
    auto offset = m_offset;
    fill_offset(counter, offset);
    counter.setValue(0);
    auto items = m_items;
    TNL::Algorithms::ParallelFor<Device>::exec(0, m_content.getSize(), [counter, hashes, items, offset] __cuda_callable__ (int i) mutable {
        #ifdef __CUDA_ARCH__
        int pos = offset[hashes[i]] + atomicAdd(&counter[hashes[i]], 1);
        atomicExch(&items[pos], i);
        #else
        int pos = offset[hashes[i]] + counter[hashes[i]]++;
        items[pos] = i;
        #endif
    });
}

template<typename Item, typename Key, typename Device>
HashGraphV2View<Item, Key, Device>::~HashGraphV2View() {}

template<typename Item, typename Key, typename Device>
bool HashGraphV2View<Item, Key, Device>::find(const Key& key, ArrayView<Item, Device> item) const {
    int hash = m_hash(key);
    int end = m_offset.getElement(hash + 1);
    auto content = m_content;
    Array<int, Device> result(1);
    auto rView = result.getView();
    rView.setValue(-1);
    auto items = m_items;
    auto _find = [content, items, item, rView, key] __cuda_callable__ (int i) mutable {
        if (content[items[i]].key == key) {
            rView[0] = items[i];
            item[0] = content[items[i]];
        }
    };
    TNL::Algorithms::ParallelFor<Device>::exec(m_offset.getElement(hash), end, _find);
    return rView.getElement(0) > -1;
}

template<typename Item, typename Key, typename Device>
void HashGraphV2View<Item, Key, Device>::fill_offset(const ArrayView<int, Device>& counter,
                                                     ArrayView<int, Device> offset) {
    auto fill = [offset, counter] __cuda_callable__ (int i) mutable {
        offset[i] = i >= counter.getSize() ? 0 : counter[i];
    };
    TNL::Algorithms::ParallelFor<Device>::exec(0, offset.getSize(), fill);
    auto reduce = [offset] __cuda_callable__ (int k, int d) mutable {
        if (k % (1 << (d + 1)) == 0)
            offset[k + (1 << (d + 1)) - 1] = offset[k + (1 << d) - 1]
                                           + offset[k + (1 << (d + 1)) - 1];
    };
    for (int d = 0; d < log2(offset.getSize() - 1); d++)
        TNL::Algorithms::ParallelFor<Device>::exec(0, offset.getSize() - 1, reduce, d);
    offset.setElement(offset.getSize() - 1, offset.getElement(offset.getSize() - 2));
    offset.setElement(offset.getSize() - 2, 0);
    auto up_sweep = [offset] __cuda_callable__ (int r, int d) mutable {
        if ((offset.getSize() - 2 - r) % (1 << d) == 0) {
            int l = r - (1 << (d - 1));
            int t = offset[r];
            offset[r] += offset[l];
            offset[l] = t;
        }
    };
    for (int d = log2(offset.getSize() - 1); d > 0; d--)
        TNL::Algorithms::ParallelFor<Device>::exec(0, offset.getSize() - 1, up_sweep, d);
}
