#include "test_map.h"

#include "CuckooHash/CuckooHashMap.h"
#include "HashGraph/HashGraphMap.h"

#include <TNL/Algorithms/ParallelFor.h>
#include <TNL/Containers/StaticArray.h>

#include <algorithm>
#include <vector>


namespace TestMap {
double g_results[6][4];
int g_row, g_col;

using Coordinates3D = TNL::Containers::StaticArray<3, int>;

template<typename K, typename V>
using PairArray = Array<Pair<K, V>, Device>;

template<typename K, typename V>
using CpuPairs = std::vector<Pair<K, V>>;


/**
 * Wrapper class for testing std::map
 */
template<typename K, typename V, typename Device>
class StdMapWrapper {
    struct Comparator {
        bool operator()(const K& v1, const K& v2) const {
            return v1 < v2;
        }
    };

  public:
    StdMapWrapper(const std::vector<Pair<K, V>>& values) {
        for (int i = 0; i < values.size(); i++)
            m_map[values[i].key] = values[i].value;        
    }

    bool find(const K& key, V& value) const {
        auto it = m_map.find(key);
        if (it == m_map.end())
            return false;
        value = it->second;
        return true;
    }

    void debug_print() const {}

  private:
    std::map<K, V, Comparator> m_map;
};


template<>
bool StdMapWrapper<Coordinates3D, int, Device>::Comparator::operator()(const Coordinates3D& v1, const Coordinates3D& v2) const {
    for (int i = 0; i < 3; i++)
        if (v1[i] != v2[i])
            return v1[i] < v2[i];
    return false;
}


template<template<typename, typename, class> class M, typename K, typename V>
M<K, V, Device> test_building(const PairArray<K, V>& pairs) {
    BEGIN_TEST(BUILDING)
    std::cerr << "\t\tInserting " << pairs.getSize() << " entries" << std::endl;
    auto table = M<K, V, Device>(pairs);
    if constexpr (std::is_same<M<K, V, Device>, CuckooHashMap<K, V, Device>>::value) {
        std::cerr << "\t\t - " << table.generations() 
                  << " generation(s) before success" << std::endl;
    }
    END_TEST
    return table;
}


template<template<typename, typename, class> class M, typename K, typename V>
M<K, V, Device> test_building(const std::vector<Pair<K, V>>& pairs) {
    BEGIN_TEST(BUILDING)
    std::cerr << "\t\tInserting " << pairs.size() << " entries" << std::endl;
    auto table = M<K, V, Device>(pairs);
    END_TEST
    return table;
}


template<template<typename, typename, class> class M, typename K, typename V>
void test_correct_query(const M<K, V, Device>& table, 
                        const CpuPairs<K, V>& cpuPairs,
                        int nQueries) {
    BEGIN_TEST(CORRECT QUERY)
    for (int i = 0; i < nQueries; i++) {
        V value;
        if (!table.find(cpuPairs[i].key, value)) {
            std::cerr << "\t\tTEST FAILED:\tKey not found, "
                      << "key: \"" << cpuPairs[i].key << "\", "
                      << "expected: \"" << cpuPairs[i].value << '"' << std::endl;
            assert(false);
        } else if (value != cpuPairs[i].value) {
            std::cerr << "\t\tTEST FAILED:\tFound wrong value, "
            << "key: \"" << cpuPairs[i].key << "\", "
            << "value: \"" << value << "\", "
            << "expected: \"" << cpuPairs[i].value << '"' << std::endl;
            assert(false);
        }
    }
    END_TEST
}

template<typename T>
T ruin(const T& value);

template<template<typename, typename, class> class M, typename K, typename V>
void test_wrong_query(const M<K, V, Device>& table, 
                      const CpuPairs<K, V>& cpuPairs,
                      int nQueries) {
    BEGIN_TEST(WRONG QUERY)
    for (int i = 0; i < nQueries; i++) {
        V value;
        auto ruined = ruin(cpuPairs[i].key);
        if (table.find(ruined, value)) {
            std::cerr << "\t\tTEST FAILED:\tFound missing key, key: \"" << cpuPairs[i].key 
            << "\" value: \"" << value << '"' << std::endl;
            assert(false);
        }
    }
    END_TEST
}

template<template<typename, typename, class> class M, typename K, typename V>
void test_random(const typename Array<Pair<K, V>, Device>::ConstViewType pairs, const CpuPairs<K, V>& cpuPairs, int nTests, double addedPart) {
    int nAdded = int(pairs.getSize() * addedPart);
    M<K, V, Device> table([&] {
        if constexpr (std::is_same<M<K, V, Device>, StdMapWrapper<K, V, Device>>::value)
            return M<K, V, Device>(std::vector<Pair<K, V>>(cpuPairs.begin(), cpuPairs.begin() + nAdded));
        else {
            auto arr = Array<Pair<K, V>, Device>(nAdded);
            auto aView = arr.getView();
            auto copy_part = [aView, pairs] __cuda_callable__ (int i) mutable {
                aView[i] = pairs[i];
            };
            TNL::Algorithms::ParallelFor<Device>::exec(0, nAdded, copy_part);
            return M<K, V, Device>(arr);
        }
    }());
    int correct = 0, wrong = 0;
    BEGIN_TEST(RANDOM)
    std::cerr << "\t\tFor " << pairs.getSize() << " entries, "
              << nAdded << '(' << addedPart * 100 << "%)"
              << " are added" << std::endl;
    for (int i = 0; i < nTests; i++) {
        int ind = rand() % pairs.getSize();
        V value;
        if (table.find(cpuPairs[ind].key, value)) {
            if (ind >= nAdded) {
                std::cerr << "\t\tTEST FAILED:\tFound missing key, "
                          << "key: \"" << cpuPairs[ind].key << "\", "
                          << "value: \"" << value << "\", "
                          << "expected missing key" << std::endl;
                assert(false);
            } else if (value != cpuPairs[ind].value) {
                std::cerr << "\t\tTEST FAILED:\tFound wrong value, "
                        << "key: \"" << cpuPairs[ind].key << "\", "
                        << "value: \"" << value << "\", "
                        << "expected: \"" << cpuPairs[ind].value << '"'
                        << std::endl;
                assert(false);
            }
            correct++;
        } else if (ind < nAdded) {
            std::cerr << "\t\tTEST FAILED:\tKey not found, "
                      << "key: \"" << cpuPairs[ind].key << "\", "
                      << "expected: \"" << cpuPairs[ind].value << '"'
                      << std::endl;
            assert(false);
        } else
            wrong++;
    }
    std::cerr << "\t\t - " << correct << " correct queries" << std::endl
              << "\t\t - " << wrong << " wrong queries" << std::endl;
    END_TEST
}

template<typename K, typename V>
PairArray<K, V> read_data(const char* fileName, CpuPairs<K, V>& cpuPairs);

template<template<typename, typename, class> class M, typename K, typename V>
void test(const char* fileName) {
    std::vector<Pair<K, V>> cpuPairs; 
    auto pairs = read_data<K, V>(fileName, cpuPairs);
    M<K, V, Device> table([&] {
        if constexpr (std::is_same<M<K, V, Device>, StdMapWrapper<K, V, Device>>::value)
            return test_building<M, K, V>(cpuPairs);
        else
            return test_building<M, K, V>(pairs);
    }());
    //table.debug_print();
    std::sort(cpuPairs.begin(), cpuPairs.end(), [](const Pair<K, V>& p1, const Pair<K, V>& p2) -> bool {
        for (int i = 0; i < 3; i++)
            if (p1.key[i] != p2.key[i])
                return p1.key[i] < p2.key[i];
        return false;
    });
    cpuPairs.erase(std::unique(cpuPairs.begin(), cpuPairs.end()), cpuPairs.end());
    test_correct_query<M, K, V>(table, cpuPairs, pairs.getSize());
    test_wrong_query<M, K, V>(table, cpuPairs, pairs.getSize());
    //test_random<M, K, V>(Array<Pair<K, V>, Device>(cpuPairs).getConstView(), cpuPairs, 100000, 0.5);

    g_row++;
    g_col = 0;
}

template<>
std::string ruin<std::string>(const std::string& value) {
    return "WRONG_" + value;
}

// template<>
// PairArray<std::string, std::string> read_data(const char* fileName, CpuPairs<std::string, std::string>& cpuPairs) {
//     std::ifstream dataFile(fileName);
//     int nLines = std::count(std::istreambuf_iterator<char>(dataFile), 
//                  std::istreambuf_iterator<char>(), '\n');
//     PairArray<std::string, std::string> lines(nLines);
//     std::string line;
//     dataFile.seekg(0);
//     for (int i = 0; i < nLines; i++) {
//         std::getline(dataFile, line);
//         std::istringstream iss(line);
//         std::string key, value;
//         std::getline(iss, key, ':');
//         std::getline(iss, value);
//         lines.setElement(i, {key, value});
//         cpuPairs.push_back({key, value});
//     }
//     return lines;
// }

template<>
Coordinates3D ruin<Coordinates3D>(const Coordinates3D& value) {
    auto wrong = Coordinates3D(value);
    // Coordinates are limited by [-1000..1000]
    wrong[0] -= 2000;
    return wrong;
}

template<>
PairArray<Coordinates3D, int> read_data(const char* fileName, CpuPairs<Coordinates3D, int>& cpuPairs) {
    std::ifstream dataFile(fileName);
    int nLines = std::count(std::istreambuf_iterator<char>(dataFile), 
                 std::istreambuf_iterator<char>(), '\n');
    PairArray<Coordinates3D, int> lines(nLines);
    std::string line;
    dataFile.seekg(0);
    for (int i = 0; i < nLines; i++) {
        std::getline(dataFile, line);
        std::istringstream iss(line);
        Coordinates3D key;
        int value;
        for (int j = 0; j < 3; j++) {
            iss >> value;
            key[j] = value;
        }
        iss >> value;
        lines.setElement(i, {key, value});
        cpuPairs.push_back({key, value});
    }
    return lines;
}

template<template<typename, typename, class> class M>
void test_class(const std::string className) {
    //std::cerr << "Testing " << className << "<std::string, std::string>:" << std::endl;
    //test<M, std::string, std::string>("data/cities.txt");

    std::cerr << "Testing " << className << "<TNL::Containers::StaticArray<3, int>, int>:" << std::endl;
    test<M, Coordinates3D, int>("data/coordinates.txt");
}


void print_results_table() {
    std::string mapClasses[] = {"std::map", "CuckooHashMap", "HashGraphV1Map"};
    std::string keyAndValue[] = {/*"<std::string, std::string>",*/ "<StaticArray<3, int>, int>"};
    std::cout << "MAP CLASS\tKEY AND VALUE\t\t\tBUILDING\tCORRECT QUERY\tWRONG QUERY\tRANDOM" << std::endl;
    int r = 0, c = 0;
    for (auto& sc : mapClasses) {
        for (auto& kv : keyAndValue) {
            std::cout << sc << '\t' << kv;
            for (int i = 0; i < 4; i++)
                std::cout << '\t' << std::setfill('0') << std::setw(8) << std::left << g_results[r][c++];
            std::cout << std::endl;
            r++;
            c = 0;
        }
    }
    std::cout << std::endl << std::endl;
}


void run() {
    test_class<StdMapWrapper>("std::map");
    test_class<CuckooHashMap>("CuckooHashMap");
    test_class<HashGraphV1Map>("HashGraphV1Map");
    test_class<HashGraphV2Map>("HashGraphV2Map");

    print_results_table();
}
}
