#include "test_set.h"

#include "CuckooHash/CuckooHashSet.h"
#include "HashGraph/HashGraphSet.h"

#include "Triangle.hpp"

#include <TNL/Algorithms/ParallelFor.h>
#include <TNL/Containers/StaticArray.h>

#include <iomanip>
#include <set>
#include <vector>


namespace TestSet {
double g_results[40][4];
int g_row, g_col;

std::ostream& operator<<(std::ostream& os, const TestSet::Triangle& t) {
    os << '[' << t.v[0] << ", " << t.v[1] << ", " << t.v[2] << ']';
    return os;
}

using Cell = TNL::Containers::StaticArray<4, int>;

/**
 * Wrapper class for testing std::set
 */
template<typename T, typename Device>
class StdSetWrapper {
    struct Comparator {
        bool operator()(const T& v1, const T& v2) const {
            return v1 < v2;
        }
    };

  public:
    StdSetWrapper(const std::vector<T>& values) :
                    m_set(values.begin(), values.end()) {}
    
    StdSetWrapper(StdSetWrapper<T, Device>&& other) : m_set(std::move(other.m_set)) {}

    bool contains(const T& value) const {
        return m_set.count(value);
    }

    int generations() { return 1; }
    std::string duplicates() { return "Unknown number of"; }

  private:
    std::set<T, Comparator> m_set;
};

template<>
bool StdSetWrapper<Cell, Device>::Comparator::operator()(const Cell& v1, const Cell& v2) const {
    for (int i = 0; i < 4; i++)
        if (v1[i] != v2[i])
            return v1[i] < v2[i];
    return false;
}


template<template<class, class> class S, typename T>
S<T, Device> test_building(const Array<T, Device>& values) {
    BEGIN_TEST(BUILDING)
    std::cerr << "\t\tInserting " << values.getSize() << " entries" << std::endl;
    S<T, Device> table(values);
    if constexpr (std::is_same<S<T, Device>, CuckooHashSet<T, Device>>::value) {
        std::cerr << "\t\t - " << table.generations() 
                  << " generation(s) before success" << std::endl;
    }
    std::cerr << "\t\t - " << table.duplicates()
              << " values were the duplicates" << std::endl;
    END_TEST
    return table;
}


template<template<class, class> class S, typename T>
S<T, Device> test_building(const std::vector<T>& values) {
    BEGIN_TEST(BUILDING)
    std::cerr << "\t\tInserting " << values.size() << " entries" << std::endl;
    auto table = S<T, Device>(values);
    END_TEST
    return table; 
}


template<template<class, class> class S, typename T>
void test_correct_query(const S<T, Device>& table, 
                        const std::vector<T>& values, 
                        int nQueries) {
    BEGIN_TEST(CORRECT QUERY)
    for (int i = 0; i < nQueries; i++) {
        if (!table.contains(values[i])) {
            std::cerr << "\t\tTEST FAILED:\tKey not found: "
                      << '"' << values[i] << '"';
            assert(false);
        }
    }
    END_TEST
}

template<typename T>
T ruin(const T& value);

template<template<class, class> class S, typename T>
void test_wrong_query(const S<T, Device>& table, 
                      const std::vector<T>& values, 
                      int nQueries) {
    BEGIN_TEST(WRONG QUERY)
    for (int i = 0; i < nQueries; i++) {
        auto ruined = ruin(values[i]);
        if (table.contains(ruined)) {
            std::cerr << "\t\tTEST FAILED:\tFound missing key: \"" << ruined << '"' << std::endl;
            assert(false);
        }
    }
    END_TEST
}


template<template<class, class> class S, typename T>
void test_random(const typename Array<T, Device>::ConstViewType values, const std::vector<T>& cpuValues, int nTests, double addedPart) {
    int nAdded = int(values.getSize() * addedPart);
    S<T, Device> table([&]() {
        if constexpr (std::is_same<S<T, Device>, StdSetWrapper<T, Device>>::value) {
            using namespace std;
            return std::vector(cpuValues.begin(), cpuValues.begin() + nAdded);
        } else {
            auto arr = Array<T, Device>(nAdded);
            auto aView = arr.getView();
            auto copy_part = [aView, values] __cuda_callable__ (int i) mutable {
                aView[i] = values[i];
            };
            TNL::Algorithms::ParallelFor<Device>::exec(0, nAdded, copy_part);
            return arr;
        }
    }());
    //auto table = build_table();
    int correct = 0, wrong = 0;
    BEGIN_TEST(RANDOM)
    std::cerr << "\t\tFor " << cpuValues.size() << " entries, "
              << nAdded << '(' << addedPart * 100 << "%)"
              << " are added" << std::endl;
    for (int i = 0; i < nTests; i++) {
        int ind = rand() % cpuValues.size();
        if (table.contains(cpuValues[ind])) {
            correct++;
            if (ind >= nAdded) {
                std::cerr << "\t\tTEST FAILED:\tFound missing key: "
                          << '"' << cpuValues[ind] << '"' << std::endl;
                assert(false);
            }
        } else if (ind < nAdded) {
            std::cerr << "\t\tTEST FAILED:\tKey not found: "
                      << '"' << cpuValues[ind] << '"' << std::endl;
            assert(false);
        } else
            wrong++;
    }
    std::cerr << "\t\t - " << correct << " correct queries" << std::endl
              << "\t\t - " << wrong << " wrong queries" << std::endl;
    END_TEST
}

template<typename T>
Array<T, Device> read_data(std::string fileName, std::vector<T>& cpu);

template<typename T>
Array<T, Device> get_data(std::string fileName, std::vector<T>& cpu) {
    static std::map<std::string, std::pair<Array<T, Device>, std::vector<T>>> dataMap;
    auto it = dataMap.find(fileName);
    if (it == dataMap.end()) {
        typename std::remove_reference<decltype(cpu)>::type cpuCopy;
        auto gpu = read_data<T>(fileName, cpuCopy);
        cpu = cpuCopy;
        dataMap[fileName] = std::pair<decltype(gpu), decltype(cpuCopy)>(gpu, cpuCopy);
    }
    auto dataPair = dataMap[fileName];
    cpu = dataPair.second;
    return Array<T, Device>(dataPair.first);
}

template<>
Array<Triangle, Device> read_data(std::string fileName, std::vector<Triangle>& cpuLines) {
    std::ifstream dataFile(fileName);
    int nLines = std::count(std::istreambuf_iterator<char>(dataFile), 
                 std::istreambuf_iterator<char>(), '\n');
    Array<Triangle, Device> gpuLines(nLines);
    std::string line;
    dataFile.seekg(0);
    std::getline(dataFile, line);
    for (int i = 0; i < nLines; i++) {
        std::getline(dataFile, line);
        std::istringstream iss(line);
        int coords[3];
        for (int j = 0; j < 4; j++)
            iss >> coords[j];
        Triangle triangle(coords);
        gpuLines.setElement(i, triangle);
        cpuLines.push_back(triangle);
    }
    return gpuLines;
}


template<>
Array<Cell, Device> read_data(std::string fileName, std::vector<Cell>& cpuLines) {
    std::ifstream dataFile(fileName);
    int nLines = std::count(std::istreambuf_iterator<char>(dataFile), 
                            std::istreambuf_iterator<char>(), '\n');
    Array<Cell, Device> gpuLines(nLines);
    std::string line;
    dataFile.seekg(0);
    std::getline(dataFile, line);
    for (int i = 0; i < nLines; i++) {
        std::getline(dataFile, line);
        std::istringstream iss(line);
        int value;
        Cell cell;
        for (int j = 0; j < 4; j++) {
            iss >> value;
            cell[j] = value;
        }
        gpuLines.setElement(i, cell);
        cpuLines.push_back(cell);
    }
    return gpuLines;
}


template<template<class, class> class S, typename T>
void test(std::string fileName) {
    std::vector<T> cpuValues;
    auto values = get_data<T>(fileName, cpuValues);
    S<T, Device> table([&] {
        if constexpr (std::is_same<S<T, Device>, StdSetWrapper<T, Device>>::value)
            return test_building<S, T>(cpuValues);
        else
            return test_building<S, T>(values);
    }());
    std::sort(cpuValues.begin(), cpuValues.end(),
    [](const T& v1, const T& v2) -> bool {
        if constexpr (std::is_same<T, Cell>::value) {
            for (int i = 0; i < 4; i++)
                if (v1[i] != v2[i])
                    return v1[i] < v2[i];
            return false;
        } else
            return v1 < v2;
    });
    cpuValues.erase(std::unique(cpuValues.begin(), cpuValues.end()), cpuValues.end());
    test_correct_query<S, T>(table, cpuValues, cpuValues.size());
    test_wrong_query<S, T>(table, cpuValues, cpuValues.size());
    //test_random<S, T>(Array<T, Device>(cpuValues).getConstView(), cpuValues, 100000, 0.5);
    g_row++;
    g_col = 0;
}


template<>
Cell ruin<Cell>(const Cell& value) {
    auto wrong = Cell(value);
    // Cell indeces are positive
    wrong[0] = -1;
    return wrong;
}


template<>
Triangle ruin<Triangle>(const Triangle& value) {
    // Triangle's vertices cannot be in one point
    int v[3] = {0, 0, 0};
    return Triangle(v);
}


template<template<class, class> class S>
void test_class(const std::string className, const std::string dataFile) {
    std::cerr << "Testing " << className << "<TNL::Containers::StaticArray<4, int>> with " << dataFile << ':' << std::endl;
    test<S, Cell>(dataFile);

    std::cerr <<  "Testing " << className << "<Triangle> with " << dataFile << ':' << std::endl;
    test<S, Triangle>(dataFile);
}


void print_results_table() {
    std::string setClasses[] = {"std::set", "CuckooHashSet", "HashGraphV1Set", "HashGraphV2Set"};
    std::string keyAndValue[] = {"<StaticArray<3, int>, int>", "<Triangle>\t\t"};
    std::cout << "DATA FILE\t\tSET CLASS\tKEY AND VALUE\t\t\tBUILDING\tCORRECT QUERY\tWRONG QUERY\tRANDOM" << std::endl;
    int r = 0, c = 0;
    for (int i = 1; i <= 5; i++) {
        for (auto& sc : setClasses) {
            for (auto& kv : keyAndValue) {
                std::cout << "data/cube1m_" << i << "_cells.txt\t";
                std::cout << sc << '\t' << kv;
                for (int i = 0; i < 4; i++)
                    std::cout << '\t' << std::setfill(' ') << std::setw(8) << std::left << g_results[r][c++];
                std::cout << std::endl;
                r++;
                c = 0;
            }
        }
    }
    std::cout << std::endl << std::endl;
}


void run() {
    for (int i = 1; i <= 5; i++) {
        std::string dataFile = "data/cube1m_" + std::to_string(i) + "_cells.txt";

        test_class<StdSetWrapper>("std::set", dataFile);
        test_class<CuckooHashSet>("CuckooHashSet", dataFile);
        test_class<HashGraphV1Set>("HashGraphV1Set", dataFile);
        test_class<HashGraphV2Set>("HashGraphV2Set", dataFile);
    }
    print_results_table();
}
}

